@@ -1,9 +1,9 @@
-/*	$Id: buffer.c,v 1.28 2020/06/19 10:25:42 kristaps Exp $ */
+/*	$Id: buffer.c,v 1.31 2021/01/22 13:03:37 kristaps Exp $ */
 /*
  * Copyright (c) 2008, Natacha Porté
  * Copyright (c) 2011, Vicent Martí
  * Copyright (c) 2014, Xavier Mendez, Devin Torres and the Hoedown authors
- * Copyright (c) 2016, Kristaps Dzonsons
+ * Copyright (c) 2016, 2021, Kristaps Dzonsons
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -36,34 +36,29 @@
 static void
 hbuf_init(struct lowdown_buf *buf, size_t unit, int buffer_free)
 {
-	assert(buf);
 
+	assert(buf != NULL);
 	buf->data = NULL;
-	buf->size = buf->asize = 0;
+	buf->size = buf->maxsize = 0;
 	buf->unit = unit;
 	buf->buffer_free = buffer_free;
 }
 
-/*
- * Clone the buffer at "buf" into the one at "v".
- * The storage of "v" is externally managed.
- * This is a deep copy.
- * Always returns a valid pointer to "v".
- */
-struct lowdown_buf *
+int
 hbuf_clone(const struct lowdown_buf *buf, struct lowdown_buf *v)
 {
 
 	v->data = NULL;
 	if (buf->size) {
-		v->data = xmalloc(buf->size);
+		if ((v->data = malloc(buf->size)) == NULL)
+			return 0;
 		memcpy(v->data, buf->data, buf->size);
 	} 
 	v->size = buf->size;
-	v->asize = buf->asize;
+	v->maxsize = buf->maxsize;
 	v->unit = buf->unit;
 	v->buffer_free = buf->buffer_free;
-	return(v);
+	return 1;
 }
 
 void
@@ -101,24 +96,17 @@ hbuf_eq(const struct lowdown_buf *buf1, const struct lowdown_buf *buf2)
 	       memcmp(buf1->data, buf2->data, buf1->size) == 0;
 }
 
-/* 
- * Allocate a new buffer.
- * FIXME: the "unit" value is stupid (see hbuf_grow()).
- * Always returns a valid pointer (ENOMEM aborts).
- */
 struct lowdown_buf *
 hbuf_new(size_t unit)
 {
 	struct lowdown_buf	*ret;
 
-	ret = xmalloc(sizeof(struct lowdown_buf));
+	if ((ret = malloc(sizeof(struct lowdown_buf))) == NULL)
+		return NULL;
 	hbuf_init(ret, unit, 1);
 	return ret;
 }
 
-/*
- * Exported function name.
- */
 struct lowdown_buf *
 lowdown_buf_new(size_t unit)
 {
@@ -126,18 +114,13 @@ lowdown_buf_new(size_t unit)
 	return hbuf_new(unit);
 }
 
-/* 
- * Free the buffer.
- * Passing NULL is a noop.
- */
 void
 hbuf_free(struct lowdown_buf *buf)
 {
-	if (NULL == buf) 
+
+	if (buf == NULL) 
 		return;
-
 	free(buf->data);
-
 	if (buf->buffer_free)
 		free(buf);
 }
@@ -149,144 +132,122 @@ lowdown_buf_free(struct lowdown_buf *buf)
 	hbuf_free(buf);
 }
 
-/* 
- * Increase the allocated size to the given value.
- * May not be NULL.
- * Always succeeds: ENOMEM will abort.
- * Note: "unit" must be defined (grow size).
- */
-void
+int
 hbuf_grow(struct lowdown_buf *buf, size_t neosz)
 {
-	size_t neoasz;
+	size_t	 neoasz;
+	void	*pp;
 
-	/* 
-	 * FIXME: this is a stupid assertion.
-	 * hbuf_new should have a default value that doesn't change
-	 * depending on whether we're a block/span, which is just a
-	 * useless micro optimisation.
-	 */
-	assert(buf && buf->unit);
-
-	if (buf->asize >= neosz)
-		return;
+	if (buf->maxsize >= neosz)
+		return 1;
 
 	/* FIXME: don't loop!  Use arithmetic! */
 
-	neoasz = buf->asize + buf->unit;
+	neoasz = buf->maxsize + buf->unit;
 	while (neoasz < neosz)
 		neoasz += buf->unit;
 
-	buf->data = xrealloc(buf->data, neoasz);
-	buf->asize = neoasz;
+	if ((pp = realloc(buf->data, neoasz)) == NULL)
+		return 0;
+	buf->data = pp;
+	buf->maxsize = neoasz;
+	return 1;
 }
 
-void
+int
 hbuf_putb(struct lowdown_buf *buf, const struct lowdown_buf *b)
 {
 
-	if (b != NULL)
-		hbuf_put(buf, b->data, b->size);
+	assert(buf != NULL && b != NULL);
+	return hbuf_put(buf, b->data, b->size);
 }
 
-/* 
- * Append raw data to a buffer.
- * May not be NULL.
- * See hbuf_grow().
- */
-void
+int
 hbuf_put(struct lowdown_buf *buf, const char *data, size_t size)
 {
-	assert(buf && buf->unit);
+	assert(buf != NULL && buf->unit);
 
 	if (data == NULL || size == 0)
-		return;
+		return 1;
 
-	if (buf->size + size > buf->asize)
-		hbuf_grow(buf, buf->size + size);
+	if (buf->size + size > buf->maxsize &&
+	    !hbuf_grow(buf, buf->size + size))
+		return 0;
 
 	memcpy(buf->data + buf->size, data, size);
 	buf->size += size;
+	return 1;
 }
 
-/* 
- * Append a nil-terminated string to a buffer.
- * Neither may be NULL.
- * See hbuf_put().
- */
-void
+int
 hbuf_puts(struct lowdown_buf *buf, const char *str)
 {
 
-	assert(buf && str);
-	hbuf_put(buf, str, strlen(str));
+	assert(buf != NULL && str != NULL);
+	return hbuf_put(buf, str, strlen(str));
 }
 
-/* 
- * Append a single char to a buffer.
- * May not be NULL.
- * See hbuf_grow().
- */
-void
+int
 hbuf_putc(struct lowdown_buf *buf, char c)
 {
 	assert(buf && buf->unit);
 
-	if (buf->size >= buf->asize)
-		hbuf_grow(buf, buf->size + 1);
+	if (buf->size >= buf->maxsize &&
+	    !hbuf_grow(buf, buf->size + 1))
+		return 0;
 
 	buf->data[buf->size] = c;
 	buf->size += 1;
+	return 1;
 }
 
-/* 
- * Read from a file and append to a buffer, until EOF or error.
- * Returns ferror(3).
- */
 int
 hbuf_putf(struct lowdown_buf *buf, FILE *file)
 {
-	assert(buf && buf->unit);
 
+	assert(buf != NULL && buf->unit);
 	while (!(feof(file) || ferror(file))) {
-		hbuf_grow(buf, buf->size + buf->unit);
-		buf->size += fread(buf->data + buf->size, 1, buf->unit, file);
+		if (!hbuf_grow(buf, buf->size + buf->unit))
+			return 0;
+		buf->size += fread(buf->data + buf->size, 
+			1, buf->unit, file);
 	}
 
-	return ferror(file);
+	return ferror(file) == 0;
 }
 
-/* 
- * Formatted printing to a buffer.
- */
-void
+int
 hbuf_printf(struct lowdown_buf *buf, const char *fmt, ...)
 {
-	va_list ap;
-	int n;
+	va_list	 ap;
+	int	 n;
 
-	assert(buf && buf->unit);
+	assert(buf != NULL && buf->unit);
 
-	if (buf->size >= buf->asize)
-		hbuf_grow(buf, buf->size + 1);
+	if (buf->size >= buf->maxsize &&
+	    !hbuf_grow(buf, buf->size + 1))
+		return 0;
 
 	va_start(ap, fmt);
-	n = vsnprintf(buf->data + buf->size, buf->asize - buf->size, fmt, ap);
+	n = vsnprintf(buf->data + buf->size,
+		buf->maxsize - buf->size, fmt, ap);
 	va_end(ap);
 
 	if (n < 0)
-		return;
-
-	if ((size_t)n >= buf->asize - buf->size) {
-		hbuf_grow(buf, buf->size + n + 1);
+		return 0;
 
+	if ((size_t)n >= buf->maxsize - buf->size) {
+		if (!hbuf_grow(buf, buf->size + n + 1))
+			return 0;
 		va_start(ap, fmt);
-		n = vsnprintf(buf->data + buf->size, buf->asize - buf->size, fmt, ap);
+		n = vsnprintf(buf->data + buf->size,
+			buf->maxsize - buf->size, fmt, ap);
 		va_end(ap);
 	}
 
 	if (n < 0)
-		return;
+		return 0;
 
 	buf->size += n;
+	return 1;
 }
