@@ -1,4 +1,4 @@
-
+/*	$Id: diff.c,v 1.41 2021/02/01 13:32:35 kristaps Exp $ */
 /*
  * Copyright (c) 2017, 2018 Kristaps Dzonsons <kristaps@bsd.lv>
  *
@@ -37,8 +37,6 @@
 #include "libdiff.h"
 #include "extern.h"
 
-#define	DEBUG 0
-
 struct	xnode {
 	char		 sig[MD5_DIGEST_STRING_LENGTH]; /* signature */
 	double		 weight; /* priority queue weight */
@@ -50,8 +48,9 @@ struct	xnode {
 
 struct	xmap {
 	struct xnode	*nodes; /* all of the nodes (dense table) */
-	size_t		 maxid; /* maximum id in nodes */
 	size_t		 maxsize; /* size of "nodes" (allocation) */
+	size_t		 maxid; /* maximum id in nodes */
+	size_t		 maxnodes; /* filled entries in "nodes" */
 	double		 maxweight; /* maximum node weight */
 };
 
@@ -72,49 +71,6 @@ struct	sesnode {
 	int		 headsp; /* whether there's leading space */
 };
 
-#if DEBUG
-static	const char *const names[LOWDOWN__MAX] = {
-	"LOWDOWN_ROOT",			/* LOWDOWN_ROOT */
-	"LOWDOWN_BLOCKCODE",            /* LOWDOWN_BLOCKCODE */
-	"LOWDOWN_BLOCKQUOTE",           /* LOWDOWN_BLOCKQUOTE */
-	"LOWDOWN_DEFINITION",		/* LOWDOWN_DEFINITION */
-	"LOWDOWN_DEFINITION_TITLE",	/* LOWDOWN_DEFINITION_TITLE */
-	"LOWDOWN_DEFINITION_DATA",	/* LOWDOWN_DEFINITION_DATA */
-	"LOWDOWN_HEADER",               /* LOWDOWN_HEADER */
-	"LOWDOWN_HRULE",                /* LOWDOWN_HRULE */
-	"LOWDOWN_LIST",                 /* LOWDOWN_LIST */
-	"LOWDOWN_LISTITEM",             /* LOWDOWN_LISTITEM */
-	"LOWDOWN_PARAGRAPH",            /* LOWDOWN_PARAGRAPH */
-	"LOWDOWN_TABLE_BLOCK",          /* LOWDOWN_TABLE_BLOCK */
-	"LOWDOWN_TABLE_HEADER",         /* LOWDOWN_TABLE_HEADER */
-	"LOWDOWN_TABLE_BODY",           /* LOWDOWN_TABLE_BODY */
-	"LOWDOWN_TABLE_ROW",            /* LOWDOWN_TABLE_ROW */
-	"LOWDOWN_TABLE_CELL",           /* LOWDOWN_TABLE_CELL */
-	"LOWDOWN_FOOTNOTES_BLOCK",      /* LOWDOWN_FOOTNOTES_BLOCK */
-	"LOWDOWN_FOOTNOTE_DEF",         /* LOWDOWN_FOOTNOTE_DEF */
-	"LOWDOWN_BLOCKHTML",            /* LOWDOWN_BLOCKHTML */
-	"LOWDOWN_LINK_AUTO",            /* LOWDOWN_LINK_AUTO */
-	"LOWDOWN_CODESPAN",             /* LOWDOWN_CODESPAN */
-	"LOWDOWN_DOUBLE_EMPHASIS",      /* LOWDOWN_DOUBLE_EMPHASIS */
-	"LOWDOWN_EMPHASIS",             /* LOWDOWN_EMPHASIS */
-	"LOWDOWN_HIGHLIGHT",            /* LOWDOWN_HIGHLIGHT */
-	"LOWDOWN_IMAGE",                /* LOWDOWN_IMAGE */
-	"LOWDOWN_LINEBREAK",            /* LOWDOWN_LINEBREAK */
-	"LOWDOWN_LINK",                 /* LOWDOWN_LINK */
-	"LOWDOWN_TRIPLE_EMPHASIS",      /* LOWDOWN_TRIPLE_EMPHASIS */
-	"LOWDOWN_STRIKETHROUGH",        /* LOWDOWN_STRIKETHROUGH */
-	"LOWDOWN_SUPERSCRIPT",          /* LOWDOWN_SUPERSCRIPT */
-	"LOWDOWN_FOOTNOTE_REF",         /* LOWDOWN_FOOTNOTE_REF */
-	"LOWDOWN_MATH_BLOCK",           /* LOWDOWN_MATH_BLOCK */
-	"LOWDOWN_RAW_HTML",             /* LOWDOWN_RAW_HTML */
-	"LOWDOWN_ENTITY",               /* LOWDOWN_ENTITY */
-	"LOWDOWN_NORMAL_TEXT",          /* LOWDOWN_NORMAL_TEXT */
-	"LOWDOWN_DOC_HEADER",           /* LOWDOWN_DOC_HEADER */
-	"LOWDOWN_META",           	/* LOWDOWN_META */
-	"LOWDOWN_DOC_FOOTER",           /* LOWDOWN_DOC_FOOTER */
-};
-#endif
-
 static void
 MD5Updatebuf(MD5_CTX *ctx, const struct lowdown_buf *v)
 {
@@ -137,43 +93,49 @@ MD5Updatev(MD5_CTX *ctx, const void *v, size_t sz)
  * Returns the weight of the node rooted at "n".
  * If "parent" is not NULL, its hash is updated with the hash computed
  * for the current "n" and its children.
+ * Return <0 on failure.
  */
 static double
 assign_sigs(MD5_CTX *parent, struct xmap *map, 
 	const struct lowdown_node *n)
 {
-	const struct lowdown_node *nn;
-	size_t		 weight = 0;
-	MD5_CTX		 ctx;
-	double		 v;
-	struct xnode	*xn;
+	const struct lowdown_node	*nn;
+	ssize_t				 weight = -1;
+	MD5_CTX				 ctx;
+	double				 v = 0.0, vv;
+	struct xnode			*xn;
+	void				*pp;
 
 	/* Get our node slot. */
 
 	if (n->id >= map->maxsize) {
-		map->nodes = xrecallocarray
-			(map->nodes, 
-			 map->maxsize, n->id + 64,
-			 sizeof(struct xnode));
+		pp = recallocarray(map->nodes, map->maxsize, 
+			n->id + 64, sizeof(struct xnode));
+		if (pp == NULL)
+			return -1.0;
+		map->nodes = pp;
 		map->maxsize = n->id + 64;
 	}
 
-	assert(n->id < map->maxsize);
 	xn = &map->nodes[n->id];
-	assert(NULL == xn->node);
-	assert(0.0 == xn->weight);
+	assert(xn->node == NULL);
+	assert(xn->weight == 0.0);
 	xn->node = n;
 	if (n->id > map->maxid)
 		map->maxid = n->id;
+	assert(map->maxid < map->maxsize);
+	map->maxnodes++;
 
 	/* Recursive step. */
 
 	MD5Init(&ctx);
 	MD5Updatev(&ctx, &n->type, sizeof(enum lowdown_rndrt));
 
-	v = 0.0;
-	TAILQ_FOREACH(nn, &n->children, entries)
-		v += assign_sigs(&ctx, map, nn);
+	TAILQ_FOREACH(nn, &n->children, entries) {
+		if ((vv = assign_sigs(&ctx, map, nn)) < 0.0)
+			return vv;
+		v += vv;
+	}
 
 	/* Re-assign "xn": child might have reallocated. */
 
@@ -219,22 +181,12 @@ assign_sigs(MD5_CTX *parent, struct xmap *map,
 		break;
 	}
 
-	switch (n->type) {
-	case LOWDOWN_BLOCKCODE:
-	case LOWDOWN_BLOCKHTML:
-	case LOWDOWN_LINK_AUTO:
-	case LOWDOWN_CODESPAN:
-	case LOWDOWN_IMAGE:
-	case LOWDOWN_RAW_HTML:
-	case LOWDOWN_NORMAL_TEXT:
-	case LOWDOWN_ENTITY:
-		assert(0.0 == xn->weight);
-		xn->weight = 1.0 + log(weight);
-		break;
-	default:
+	/* Weight can be zero if text size is zero. */
+
+	if (weight >= 0)
+		xn->weight = 1.0 + (weight == 0 ? 0.0 : log(weight));
+	else
 		xn->weight += 1.0;
-		break;
-	}
 
 	/*
 	 * Augment our signature from our attributes.
@@ -314,14 +266,17 @@ assign_sigs(MD5_CTX *parent, struct xmap *map,
 
 	MD5End(&ctx, xn->sig);
 
-	if (NULL != parent)
+	if (parent != NULL)
 		MD5Update(parent, (uint8_t *)xn->sig, 
 			MD5_DIGEST_STRING_LENGTH - 1);
 
 	if (xn->weight > map->maxweight)
 		map->maxweight = xn->weight;
 
-	return(xn->weight);
+	assert(isfinite(xn->weight));
+	assert(isnormal(xn->weight));
+	assert(xn->weight > 0.0);
+	return xn->weight;
 }
 
 /*
@@ -329,43 +284,47 @@ assign_sigs(MD5_CTX *parent, struct xmap *map,
  * Priority is given to weights; and if weights are equal, then
  * proximity to the parse root given by a pre-order identity.
  * FIXME: use a priority heap.
+ * Return zero on failure, non-zero on success.
  */
-static void
+static int
 pqueue(const struct lowdown_node *n, 
 	struct xmap *map, struct pnodeq *pq)
 {
 	struct pnode	*p, *pp;
 	struct xnode	*xnew, *xold;
 
-	p = xmalloc(sizeof(struct pnode));
+	if ((p = malloc(sizeof(struct pnode))) == NULL)
+		return 0;
 	p->node = n;
 
 	xnew = &map->nodes[n->id];
-	assert(NULL != xnew->node);
+	assert(xnew != NULL);
+	assert(xnew->node != NULL);
 
 	TAILQ_FOREACH(pp, pq, entries) {
 		xold = &map->nodes[pp->node->id];
-		assert(NULL != xold->node);
+		assert(xold->node != NULL);
 		if (xnew->weight >= xold->weight)
 			break;
 	}
-	if (NULL == pp) {
+	if (pp == NULL) {
 		TAILQ_INSERT_TAIL(pq, p, entries);
-		return;
+		return 1;
 	} else if (xnew->weight > xold->weight) {
 		TAILQ_INSERT_BEFORE(pp, p, entries);
-		return;
+		return 1;
 	}
 
-	for (; NULL != pp; pp = TAILQ_NEXT(pp, entries)) {
+	for (; pp != NULL; pp = TAILQ_NEXT(pp, entries)) {
 		assert(p->node->id != pp->node->id);
 		if (p->node->id < pp->node->id)
 			break;
 	}
-	if (NULL == pp) 
+	if (pp == NULL) 
 		TAILQ_INSERT_TAIL(pq, p, entries);
 	else
 		TAILQ_INSERT_BEFORE(pp, p, entries);
+	return 1;
 }
 
 /*
@@ -382,24 +341,24 @@ optimality(struct xnode *xnew, struct xmap *xnewmap,
 
 	/* Height: log(n) * W/W_0 or at least 1. */
 
-	d = ceil(log(xnewmap->maxid) * 
+	d = ceil(log(xnewmap->maxnodes) * 
 		xnew->weight / xnewmap->maxweight);
 
-	if (0 == d)
+	if (d == 0)
 		d = 1;
 	
 	/* FIXME: are we supposed to bound to "d"? */
 
-	while (NULL != xnew->node->parent &&
-	       NULL != xold->node->parent && i < d) {
+	while (xnew->node->parent != NULL &&
+	       xold->node->parent != NULL && i < d) {
 		xnew = &xnewmap->nodes[xnew->node->parent->id];
 		xold = &xoldmap->nodes[xold->node->parent->id];
-		if (NULL != xnew->match && xnew->match == xold->node) 
+		if (xnew->match != NULL && xnew->match == xold->node) 
 			opt++;
 		i++;
 	}
 
-	return(opt);
+	return opt;
 }
 
 /*
@@ -419,13 +378,12 @@ candidate(struct xnode *xnew, struct xmap *xnewmap,
 	size_t		 opt;
 	long long	 dnew, dold;
 
-	assert(NULL != xnew->node);
-	assert(NULL != xold->node);
+	assert(xnew->node != NULL);
+	assert(xold->node != NULL);
 
-	if (NULL == xnew->optmatch) {
+	if (xnew->optmatch == NULL) {
 		xnew->optmatch = xold->node;
-		xnew->opt = optimality
-			(xnew, xnewmap, xold, xoldmap);
+		xnew->opt = optimality(xnew, xnewmap, xold, xoldmap);
 		return;
 	}
 
@@ -462,22 +420,40 @@ match_eq(const struct lowdown_node *n1,
 {
 
 	if (n1->type != n2->type)
-		return(0);
+		return 0;
 
-	if (LOWDOWN_LINK == n1->type) {
-		/*
-		 * Links have both contained nodes (for the alt text,
-		 * which can be nested) and also attributes.
-		 */
-		if ( ! hbuf_eq(&n1->rndr_link.link,
-			       &n2->rndr_link.link))
-			return(0);
-		if ( ! hbuf_eq(&n1->rndr_link.title,
-			       &n2->rndr_link.title))
-			return(0);
+	switch (n1->type) {
+	case LOWDOWN_LINK:
+		if (!hbuf_eq
+		    (&n1->rndr_link.link, &n2->rndr_link.link))
+			return 0;
+		if (!hbuf_eq
+		    (&n1->rndr_link.title, &n2->rndr_link.title))
+			return 0;
+		break;
+	case LOWDOWN_HEADER:
+		if (n1->rndr_header.level !=
+		    n2->rndr_header.level)
+			return 0;
+		break;
+	case LOWDOWN_FOOTNOTE_DEF:
+		if (n1->rndr_footnote_def.num !=
+		    n2->rndr_footnote_def.num)
+			return 0;
+		break;
+	case LOWDOWN_LISTITEM:
+		if (n1->rndr_listitem.num !=
+		    n2->rndr_listitem.num)
+			return 0;
+		if (n1->rndr_listitem.flags !=
+		    n2->rndr_listitem.flags)
+			return 0;
+		break;
+	default:
+		break;
 	}
 
-	return(1);
+	return 1;
 }
 
 /*
@@ -487,11 +463,10 @@ static int
 match_singleton(const struct lowdown_node *n)
 {
 
-	if (NULL == n->parent)
-		return(1);
-
-	return(TAILQ_NEXT(n, entries) == 
-	       TAILQ_PREV(n, lowdown_nodeq, entries));
+	if (n->parent == NULL)
+		return 1;
+	return TAILQ_NEXT(n, entries) == 
+	       TAILQ_PREV(n, lowdown_nodeq, entries);
 }
 
 /*
@@ -510,20 +485,19 @@ match_up(struct xnode *xnew, struct xmap *xnewmap,
 
 	/* Height: log(n) * W/W_0 or at least 1. */
 
-	d = ceil(log(xnewmap->maxid) * 
+	d = ceil(log(xnewmap->maxnodes) * 
 		xnew->weight / xnewmap->maxweight);
-	if (0 == d)
+	if (d == 0)
 		d = 1;
 
-	while (NULL != xnew->node->parent &&
-	       NULL != xold->node->parent && i < d) {
+	while (xnew->node->parent != NULL &&
+	       xold->node->parent != NULL && i < d) {
 		/* Are the "labels" the same? */
-		if ( ! match_eq
-		    (xnew->node->parent, xold->node->parent))
+		if (!match_eq(xnew->node->parent, xold->node->parent))
 			break;
 		xnew = &xnewmap->nodes[xnew->node->parent->id];
 		xold = &xoldmap->nodes[xold->node->parent->id];
-		if (NULL != xold->match || NULL != xnew->match)
+		if (xold->match != NULL || xnew->match != NULL)
 			break;
 		xnew->match = xold->node;
 		xold->match = xnew->node;
@@ -538,17 +512,16 @@ match_up(struct xnode *xnew, struct xmap *xnewmap,
 	 * This is an extension of the algorithm.
 	 */
 
-	while (NULL != xnew->node->parent &&
-	       NULL != xold->node->parent) {
-		if ( ! match_singleton(xnew->node) ||
-		     ! match_singleton(xold->node))
+	while (xnew->node->parent != NULL &&
+	       xold->node->parent != NULL) {
+		if (!match_singleton(xnew->node) ||
+		    !match_singleton(xold->node))
 			break;
-		if ( ! match_eq
-		    (xnew->node->parent, xold->node->parent))
+		if (!match_eq(xnew->node->parent, xold->node->parent))
 			break;
 		xnew = &xnewmap->nodes[xnew->node->parent->id];
 		xold = &xoldmap->nodes[xold->node->parent->id];
-		if (NULL != xold->match || NULL != xnew->match)
+		if (xold->match != NULL || xnew->match != NULL)
 			break;
 		xnew->match = xold->node;
 		xold->match = xnew->node;
@@ -566,8 +539,23 @@ match_down(struct xnode *xnew, struct xmap *xnewmap,
 {
 	struct lowdown_node *nnew, *nold;
 
-	assert(NULL == xnew->match);
-	assert(NULL == xold->match);
+	/* 
+	 * If we're matching into a component that has already been
+	 * matched, we're in the subtree proper (the subtree root is
+	 * checked that it's not already matched) and the fact that this
+	 * is within a match indicates we're more the "larger" of the
+	 * matches, so unset its match status.
+	 */
+
+	if (xold->match != NULL) {
+		assert(xold->node == 
+			xnewmap->nodes[xold->match->id].match);
+		xnewmap->nodes[xold->match->id].match = NULL;
+		xold->match = NULL;
+	}
+
+	assert(xnew->match == NULL);
+	assert(xold->match == NULL);
 
 	xnew->match = xold->node;
 	xold->match = xnew->node;
@@ -575,7 +563,7 @@ match_down(struct xnode *xnew, struct xmap *xnewmap,
 	nnew = TAILQ_FIRST(&xnew->node->children);
 	nold = TAILQ_FIRST(&xold->node->children);
 
-	while (NULL != nnew) {
+	while (nnew != NULL) {
 		assert(NULL != nold);
 		xnew = &xnewmap->nodes[nnew->id];
 		xold = &xoldmap->nodes[nold->id];
@@ -583,6 +571,7 @@ match_down(struct xnode *xnew, struct xmap *xnewmap,
 		nnew = TAILQ_NEXT(nnew, entries);
 		nold = TAILQ_NEXT(nold, entries);
 	}
+	assert(nold == NULL);
 }
 
 /*
@@ -597,15 +586,18 @@ static struct lowdown_node *
 node_clone(const struct lowdown_node *v, size_t id)
 {
 	struct lowdown_node	*n;
+	int			 rc = 1;
+
+	if ((n = calloc(1, sizeof(struct lowdown_node))) == NULL)
+		return NULL;
 
-	n = xcalloc(1, sizeof(struct lowdown_node));
 	TAILQ_INIT(&n->children);
 	n->type = v->type;
 	n->id = id;
 
 	switch (n->type) {
 	case LOWDOWN_META:
-		hbuf_clone(&v->rndr_meta.key,
+		rc = hbuf_clone(&v->rndr_meta.key, 
 			&n->rndr_meta.key);
 		break;
 	case LOWDOWN_LIST:
@@ -619,38 +611,38 @@ node_clone(const struct lowdown_node *v, size_t id)
 		n->rndr_header.level = v->rndr_header.level;
 		break;
 	case LOWDOWN_NORMAL_TEXT:
-		hbuf_clone(&v->rndr_normal_text.text,
+		rc = hbuf_clone(&v->rndr_normal_text.text,
 			&n->rndr_normal_text.text);
 		break;
 	case LOWDOWN_ENTITY:
-		hbuf_clone(&v->rndr_entity.text,
+		rc = hbuf_clone(&v->rndr_entity.text,
 			&n->rndr_entity.text);
 		break;
 	case LOWDOWN_LINK_AUTO:
-		hbuf_clone(&v->rndr_autolink.link,
-			&n->rndr_autolink.link);
-		hbuf_clone(&v->rndr_autolink.text,
+		rc = hbuf_clone(&v->rndr_autolink.link,
+			&n->rndr_autolink.link) &&
+		     hbuf_clone(&v->rndr_autolink.text,
 			&n->rndr_autolink.text);
 		n->rndr_autolink.type = v->rndr_autolink.type;
 		break;
 	case LOWDOWN_RAW_HTML:
-		hbuf_clone(&v->rndr_raw_html.text,
+		rc = hbuf_clone(&v->rndr_raw_html.text,
 			&n->rndr_raw_html.text);
 		break;
 	case LOWDOWN_LINK:
-		hbuf_clone(&v->rndr_link.link,
-			&n->rndr_link.link);
-		hbuf_clone(&v->rndr_link.title,
+		rc = hbuf_clone(&v->rndr_link.link,
+			&n->rndr_link.link) &&
+		     hbuf_clone(&v->rndr_link.title,
 			&n->rndr_link.title);
 		break;
 	case LOWDOWN_BLOCKCODE:
-		hbuf_clone(&v->rndr_blockcode.text,
-			&n->rndr_blockcode.text);
-		hbuf_clone(&v->rndr_blockcode.lang,
+		rc = hbuf_clone(&v->rndr_blockcode.text,
+			&n->rndr_blockcode.text) &&
+		     hbuf_clone(&v->rndr_blockcode.lang,
 			&n->rndr_blockcode.lang);
 		break;
 	case LOWDOWN_CODESPAN:
-		hbuf_clone(&v->rndr_codespan.text,
+		rc = hbuf_clone(&v->rndr_codespan.text,
 			&n->rndr_codespan.text);
 		break;
 	case LOWDOWN_TABLE_HEADER:
@@ -666,13 +658,13 @@ node_clone(const struct lowdown_node *v, size_t id)
 		/* Don't use footnote number: mutable. */
 		break;
 	case LOWDOWN_IMAGE:
-		hbuf_clone(&v->rndr_image.link,
-			&n->rndr_image.link);
-		hbuf_clone(&v->rndr_image.title,
-			&n->rndr_image.title);
-		hbuf_clone(&v->rndr_image.dims,
-			&n->rndr_image.dims);
-		hbuf_clone(&v->rndr_image.alt,
+		rc = hbuf_clone(&v->rndr_image.link,
+			&n->rndr_image.link) &&
+		     hbuf_clone(&v->rndr_image.title,
+			&n->rndr_image.title) &&
+		     hbuf_clone(&v->rndr_image.dims,
+			&n->rndr_image.dims) &&
+		     hbuf_clone(&v->rndr_image.alt,
 			&n->rndr_image.alt);
 		break;
 	case LOWDOWN_MATH_BLOCK:
@@ -680,14 +672,19 @@ node_clone(const struct lowdown_node *v, size_t id)
 			v->rndr_math.blockmode;
 		break;
 	case LOWDOWN_BLOCKHTML:
-		hbuf_clone(&v->rndr_blockhtml.text,
+		rc = hbuf_clone(&v->rndr_blockhtml.text,
 			&n->rndr_blockhtml.text);
 		break;
 	default:
 		break;
 	}
 
-	return(n);
+	if (!rc) {
+		lowdown_node_free(n);
+		n = NULL;
+	}
+
+	return n;
 }
 
 /*
@@ -701,15 +698,20 @@ node_clonetree(const struct lowdown_node *v, size_t *id)
 	struct lowdown_node *n, *nn;
 	const struct lowdown_node *vv;
 
-	n = node_clone(v, *id++);
+	if ((n = node_clone(v, *id++)) == NULL)
+		return NULL;
 
 	TAILQ_FOREACH(vv, &v->children, entries) {
-		nn = node_clonetree(vv, id);
-		nn->parent = n;
+		if ((nn = node_clonetree(vv, id)) == NULL)
+			goto out;
 		TAILQ_INSERT_TAIL(&n->children, nn, entries);
+		nn->parent = n;
 	}
 
-	return(n);
+	return n;
+out:
+	lowdown_node_free(n);
+	return NULL;
 }
 
 /*
@@ -733,10 +735,10 @@ node_countwords(const struct lowdown_node *n)
 	/* First go through word, then trailing space. */
 
 	while (i < sz) {
-		assert( ! isspace((unsigned char)cp[i]));
+		assert(!isspace((unsigned char)cp[i]));
 		words++;
 		while (i < sz &&
-		       ! isspace((unsigned char)cp[i]))
+		       !isspace((unsigned char)cp[i]))
 			i++;
 		while (i < sz && 
 		       isspace((unsigned char)cp[i]))
@@ -749,8 +751,9 @@ node_countwords(const struct lowdown_node *n)
 /*
  * Like node_countwords(), except dupping individual words into a
  * structure.
+ * Return zero on failure (memory), non-zero on success.
  */
-static void
+static int
 node_tokenise(const struct lowdown_node *n, 
 	struct sesnode *toks, size_t toksz, char **savep)
 {
@@ -759,11 +762,13 @@ node_tokenise(const struct lowdown_node *n,
 
 	*savep = NULL;
 
-	if (0 == toksz)
-		return;
+	if (toksz == 0)
+		return 1;
 
 	sz = n->rndr_normal_text.text.size;
-	*savep = cp = xmalloc(sz + 1);
+	*savep = cp = malloc(sz + 1);
+	if (cp == NULL)
+		return 0;
 	memcpy(cp, n->rndr_normal_text.text.data, sz);
 	cp[sz] = '\0';
 
@@ -780,11 +785,11 @@ node_tokenise(const struct lowdown_node *n,
 
 	while (i < sz) {
 		assert(words < toksz);
-		assert( ! isspace((unsigned char)cp[i]));
+		assert(!isspace((unsigned char)cp[i]));
 		toks[words].buf = &cp[i];
 		toks[words].bufsz = 0;
 		while (i < sz &&
-		       ! isspace((unsigned char)cp[i])) {
+		       !isspace((unsigned char)cp[i])) {
 			toks[words].bufsz++;
 			i++;
 		}
@@ -798,6 +803,7 @@ node_tokenise(const struct lowdown_node *n,
 		       isspace((unsigned char)cp[i]))
 			i++;
 	}
+	return 1;
 }
 
 static int
@@ -810,53 +816,77 @@ node_word_cmp(const void *p1, const void *p2)
 	return 0 == strncmp(l1->buf, l2->buf, l1->bufsz);
 }
 
-static void
+/*
+ * Return zero on failure (memory), non-zero on success.
+ */
+static int
 node_lcs(const struct lowdown_node *nold,
 	const struct lowdown_node *nnew,
 	struct lowdown_node *n, size_t *id)
 {
-	const struct sesnode *tmp;
-	struct lowdown_node *nn;
-	struct sesnode	*newtok, *oldtok;
-	char		*newtokbuf, *oldtokbuf;
-	size_t		 i, newtoksz, oldtoksz;
-	struct diff	 d;
+	const struct sesnode	*tmp;
+	struct lowdown_node	*nn;
+	struct sesnode		*newtok = NULL, *oldtok = NULL;
+	char			*newtokbuf = NULL, *oldtokbuf = NULL;
+	size_t			 i, newtoksz, oldtoksz;
+	struct diff		 d;
+	int			 rc = 0;
+
+	memset(&d, 0, sizeof(struct diff));
 
 	newtoksz = node_countwords(nnew);
 	oldtoksz = node_countwords(nold);
 
-	newtok = xcalloc(newtoksz, sizeof(struct sesnode));
-	oldtok = xcalloc(oldtoksz, sizeof(struct sesnode));
+	newtok = calloc(newtoksz, sizeof(struct sesnode));
+	if (newtok == NULL)
+		goto out;
+	oldtok = calloc(oldtoksz, sizeof(struct sesnode));
+	if (oldtok == NULL)
+		goto out;
 
-	node_tokenise(nnew, newtok, newtoksz, &newtokbuf);
-	node_tokenise(nold, oldtok, oldtoksz, &oldtokbuf);
+	if (!node_tokenise(nnew, newtok, newtoksz, &newtokbuf))
+		goto out;
+	if (!node_tokenise(nold, oldtok, oldtoksz, &oldtokbuf))
+		goto out;
 
-	diff(&d, node_word_cmp, sizeof(struct sesnode), 
-		oldtok, oldtoksz, newtok, newtoksz);
+	if (!diff(&d, node_word_cmp, sizeof(struct sesnode), 
+	    oldtok, oldtoksz, newtok, newtoksz))
+		goto out;
 
 	for (i = 0; i < d.sessz; i++) {
 		tmp = d.ses[i].e;
 
 		if (tmp->headsp) {
-			nn = xcalloc(1, sizeof(struct lowdown_node));
-			TAILQ_INIT(&nn->children);
+			nn = calloc(1, sizeof(struct lowdown_node));
+			if (nn == NULL)
+				goto out;
 			TAILQ_INSERT_TAIL(&n->children, nn, entries);
+			TAILQ_INIT(&nn->children);
+
 			nn->type = LOWDOWN_NORMAL_TEXT;
 			nn->id = (*id)++;
 			nn->parent = n;
 			nn->rndr_normal_text.text.size = 1;
-			nn->rndr_normal_text.text.data = xstrdup(" ");
+			nn->rndr_normal_text.text.data = strdup(" ");
+			if (nn->rndr_normal_text.text.data == NULL)
+				goto out;
 		}
 
-		nn = xcalloc(1, sizeof(struct lowdown_node));
-		TAILQ_INIT(&nn->children);
+		nn = calloc(1, sizeof(struct lowdown_node));
+		if (nn == NULL)
+			goto out;
 		TAILQ_INSERT_TAIL(&n->children, nn, entries);
+		TAILQ_INIT(&nn->children);
+
 		nn->type = LOWDOWN_NORMAL_TEXT;
 		nn->id = (*id)++;
 		nn->parent = n;
 		nn->rndr_normal_text.text.size = tmp->bufsz;
 		nn->rndr_normal_text.text.data = 
-			xcalloc(1, tmp->bufsz + 1);
+			calloc(1, tmp->bufsz + 1);
+		if (nn->rndr_normal_text.text.data == NULL)
+			goto out;
+
 		memcpy(nn->rndr_normal_text.text.data,
 			tmp->buf, tmp->bufsz);
 		nn->chng = DIFF_DELETE == d.ses[i].type ?
@@ -866,23 +896,30 @@ node_lcs(const struct lowdown_node *nold,
 			LOWDOWN_CHNG_NONE;
 
 		if (tmp->tailsp) {
-			nn = xcalloc(1, sizeof(struct lowdown_node));
-			TAILQ_INIT(&nn->children);
+			nn = calloc(1, sizeof(struct lowdown_node));
+			if (nn == NULL)
+				goto out;
 			TAILQ_INSERT_TAIL(&n->children, nn, entries);
+			TAILQ_INIT(&nn->children);
 			nn->type = LOWDOWN_NORMAL_TEXT;
 			nn->id = (*id)++;
 			nn->parent = n;
 			nn->rndr_normal_text.text.size = 1;
-			nn->rndr_normal_text.text.data = xstrdup(" ");
+			nn->rndr_normal_text.text.data = strdup(" ");
+			if (nn->rndr_normal_text.text.data == NULL)
+				goto out;
 		}
 	}
 
+	rc = 1;
+out:
 	free(d.ses);
 	free(d.lcs);
 	free(newtok);
 	free(oldtok);
 	free(newtokbuf);
 	free(oldtokbuf);
+	return rc;
 }
 
 /*
@@ -901,31 +938,31 @@ node_merge(const struct lowdown_node *nold,
 	const struct xmap *xnewmap,
 	size_t *id)
 {
-	const struct xnode *xnew, *xold;
-	struct lowdown_node *n, *nn;
-	const struct lowdown_node *nnold;
+	const struct xnode		*xnew, *xold;
+	struct lowdown_node		*n, *nn;
+	const struct lowdown_node	*nnold;
 
 	/* 
 	 * Invariant: the current nodes are matched.
 	 * Start by putting that node into the current output.
 	 */
 
-	assert(NULL != nnew && NULL != nold);
+	assert(nnew != NULL && nold != NULL );
 	xnew = &xnewmap->nodes[nnew->id];
 	xold = &xoldmap->nodes[nold->id];
-	assert(NULL != xnew->match);
-	assert(NULL != xold->match);
-
+	assert(xnew->match != NULL);
+	assert(xold->match != NULL);
 	assert(xnew->match == xold->node);
 
-	n = node_clone(nnew, (*id)++);
+	if ((n = node_clone(nnew, (*id)++)) == NULL)
+		goto err;
 
 	/* Now walk through the children on both sides. */
 
 	nold = TAILQ_FIRST(&nold->children);
 	nnew = TAILQ_FIRST(&nnew->children);
 
-	while (NULL != nnew) {
+	while (nnew != NULL) {
 		/* 
 		 * Begin by flushing out all of the nodes that have been
 		 * deleted from the old tree at this level.
@@ -934,15 +971,16 @@ node_merge(const struct lowdown_node *nold,
 		 * new tree (not necessarily at this level, though).
 		 */
 
-		while (NULL != nold) {
+		while (nold != NULL) {
 			xold = &xoldmap->nodes[nold->id];
-			if (NULL != xold->match ||
+			if (xold->match != NULL ||
 			    LOWDOWN_NORMAL_TEXT == nold->type)
 				break;
-			nn = node_clonetree(nold, id);
+			if ((nn = node_clonetree(nold, id)) == NULL)
+				goto err;
+			TAILQ_INSERT_TAIL(&n->children, nn, entries);
 			nn->parent = n;
 			nn->chng = LOWDOWN_CHNG_DELETE;
-			TAILQ_INSERT_TAIL(&n->children, nn, entries);
 			nold = TAILQ_NEXT(nold, entries);
 		}
 
@@ -953,12 +991,13 @@ node_merge(const struct lowdown_node *nold,
 		 * (not necessarily at this level) with the old.
 		 */
 
-		while (NULL != nnew) {
+		while (nnew != NULL) {
 			xnew = &xnewmap->nodes[nnew->id];
-			if (NULL != xnew->match ||
+			if (xnew->match != NULL ||
 			    LOWDOWN_NORMAL_TEXT == nnew->type)
 				break;
-			nn = node_clonetree(nnew, id);
+			if ((nn = node_clonetree(nnew, id)) == NULL)
+				goto err;
 			TAILQ_INSERT_TAIL(&n->children, nn, entries);
 			nn->parent = n;
 			nn->chng = LOWDOWN_CHNG_INSERT;
@@ -971,32 +1010,35 @@ node_merge(const struct lowdown_node *nold,
 		 * This is an extension of the BULD algorithm.
 		 */
 
-		if (NULL != nold && NULL != nnew &&
-		    LOWDOWN_NORMAL_TEXT == nold->type &&
-		    NULL == xold->match &&
-		    LOWDOWN_NORMAL_TEXT == nnew->type &&
-		    NULL == xnew->match) {
-			node_lcs(nold, nnew, n, id);
+		if (nold != NULL && nnew != NULL &&
+		    nold->type == LOWDOWN_NORMAL_TEXT &&
+		    xold->match == NULL &&
+		    nnew->type == LOWDOWN_NORMAL_TEXT &&
+		    xnew->match == NULL) {
+			if (!node_lcs(nold, nnew, n, id))
+				goto err;
 			nold = TAILQ_NEXT(nold, entries);
 			nnew = TAILQ_NEXT(nnew, entries);
 		}
 
-		while (NULL != nold) {
+		while (nold != NULL) {
 			xold = &xoldmap->nodes[nold->id];
-			if (NULL != xold->match)
+			if (xold->match != NULL)
 				break;
-			nn = node_clonetree(nold, id);
+			if ((nn = node_clonetree(nold, id)) == NULL)
+				goto err;
+			TAILQ_INSERT_TAIL(&n->children, nn, entries);
 			nn->parent = n;
 			nn->chng = LOWDOWN_CHNG_DELETE;
-			TAILQ_INSERT_TAIL(&n->children, nn, entries);
 			nold = TAILQ_NEXT(nold, entries);
 		}
 
-		while (NULL != nnew) {
+		while ( nnew != NULL) {
 			xnew = &xnewmap->nodes[nnew->id];
-			if (NULL != xnew->match)
+			if (xnew->match != NULL)
 				break;
-			nn = node_clonetree(nnew, id);
+			if ((nn = node_clonetree(nnew, id)) == NULL)
+				goto err;
 			TAILQ_INSERT_TAIL(&n->children, nn, entries);
 			nn->parent = n;
 			nn->chng = LOWDOWN_CHNG_INSERT;
@@ -1005,7 +1047,7 @@ node_merge(const struct lowdown_node *nold,
 
 		/* Nothing more to do at this level? */
 
-		if (NULL == nnew)
+		if (nnew == NULL)
 			break;
 
 		/*
@@ -1018,11 +1060,11 @@ node_merge(const struct lowdown_node *nold,
 		 */
 
 		xnew = &xnewmap->nodes[nnew->id];
-		assert(NULL != xnew->match);
+		assert(xnew->match != NULL);
 
 		/* Scan ahead to find a matching old. */
 		
-		for (nnold = nold; NULL != nnold ; ) {
+		for (nnold = nold; nnold != NULL ; ) {
 			xold = &xoldmap->nodes[nnold->id];
 			if (xnew->node == xold->match) 
 				break;
@@ -1035,8 +1077,9 @@ node_merge(const struct lowdown_node *nold,
 		 * somewhere else in the tree.
 		 */
 
-		if (NULL == nnold) {
-			nn = node_clonetree(nnew, id);
+		if (nnold == NULL) {
+			if ((nn = node_clonetree(nnew, id)) == NULL)
+				goto err;
 			TAILQ_INSERT_TAIL(&n->children, nn, entries);
 			nn->parent = n;
 			nn->chng = LOWDOWN_CHNG_INSERT;
@@ -1046,18 +1089,19 @@ node_merge(const struct lowdown_node *nold,
 
 		/* Match found: flush old nodes til the match. */
 
-		while (NULL != nold) {
+		while (nold != NULL) {
 			xold = &xoldmap->nodes[nold->id];
 			if (xnew->node == xold->match) 
 				break;
-			nn = node_clonetree(nold, id);
+			if ((nn = node_clonetree(nold, id)) == NULL)
+				goto err;
 			TAILQ_INSERT_TAIL(&n->children, nn, entries);
 			nn->parent = n;
 			nn->chng = LOWDOWN_CHNG_DELETE;
 			nold = TAILQ_NEXT(nold, entries);
 		}
 
-		assert(NULL != nold);
+		assert(nold != NULL);
 
 		/*
 		 * Now we're in lock-step.
@@ -1066,8 +1110,10 @@ node_merge(const struct lowdown_node *nold,
 		 */
 
 		nn = node_merge(nold, xoldmap, nnew, xnewmap, id);
-		nn->parent = n;
+		if (nn == NULL)
+			goto err;
 		TAILQ_INSERT_TAIL(&n->children, nn, entries);
+		nn->parent = n;
 
 		nold = TAILQ_NEXT(nold, entries);
 		nnew = TAILQ_NEXT(nnew, entries);
@@ -1075,8 +1121,9 @@ node_merge(const struct lowdown_node *nold,
 
 	/* Flush remaining old nodes. */
 
-	while (NULL != nold) {
-		nn = node_clonetree(nold, id);
+	while (nold != NULL) {
+		if ((nn = node_clonetree(nold, id)) == NULL)
+			goto err;
 		TAILQ_INSERT_TAIL(&n->children, nn, entries);
 		nn->parent = n;
 		nn->chng = LOWDOWN_CHNG_DELETE;
@@ -1084,33 +1131,11 @@ node_merge(const struct lowdown_node *nold,
 	}
 
 	return n;
+err:
+	lowdown_node_free(n);
+	return NULL;
 }
 
-#if DEBUG
-static void
-node_print(const struct lowdown_node *n, const struct xmap *map, size_t ind)
-{
-	const struct lowdown_node *nn;
-	const struct xnode *xn;
-	size_t	 i;
-
-	xn = &map->nodes[n->id];
-	for (i = 0; i < ind; i++)
-		fputc(' ', stderr);
-
-	fprintf(stderr, "%zu:%s (%g)", n->id,names[n->type], xn->weight);
-	if (xn->match) {
-		fprintf(stderr, " -> %zu", xn->match->id);
-	}
-
-	fputc('\n', stderr);
-
-
-	TAILQ_FOREACH(nn, &n->children, entries)
-		node_print(nn, map, ind + 1);
-}
-#endif
-
 /*
  * Optimise from top down.
  * This works by selecting matching non-terminal nodes, both adjacent
@@ -1122,68 +1147,61 @@ static void
 node_optimise_topdown(const struct lowdown_node *n, 
 	struct xmap *newmap, struct xmap *oldmap)
 {
-	struct xnode	*xn, *xmatch, *xnchild, *xmchild,
-			*xnnext, *xmnext;
-	const struct lowdown_node 
-			*match, *nchild, *mchild, *nnext, 
-			*mnext;
+	struct xnode			*xn, *xmatch, *xnchild, 
+					*xmchild, *xnnext, *xmnext;
+	const struct lowdown_node	*match, *nchild, *mchild, 
+	      				*nnext, *mnext;
 
 	if (TAILQ_EMPTY(&n->children))
 		return;
 
 	xn = &newmap->nodes[n->id];
-	assert(NULL != xn);
+	assert(xn != NULL);
 
-	if (NULL == (match = xn->match))
+	if ((match = xn->match) == NULL)
 		return;
 
 	xmatch = &oldmap->nodes[match->id];
-	assert(NULL != xmatch);
+	assert(xmatch != NULL);
 
 	TAILQ_FOREACH(nchild, &n->children, entries) {
 		if (TAILQ_EMPTY(&nchild->children))
 			continue;
 		xnchild = &newmap->nodes[nchild->id];
-		assert(NULL != xnchild);
-		if (NULL == (mchild = xnchild->match))
+		assert(xnchild != NULL);
+		if ((mchild = xnchild->match) == NULL)
 			continue;
 		if (mchild->parent->id != match->id)
 			continue;
 		xmchild = &oldmap->nodes[mchild->id];
-		assert(NULL != xmchild);
+		assert(xmchild != NULL);
 
 		/* 
 		 * Do we have a non-terminal sibling after us without a
 		 * match? 
 		 */
 
-		if (NULL == (nnext = TAILQ_NEXT(nchild, entries)))
+		if ((nnext = TAILQ_NEXT(nchild, entries)) == NULL)
 			continue;
 		if (TAILQ_EMPTY(&nnext->children))
 			continue;
 		xnnext = &newmap->nodes[nnext->id];
-		assert(NULL != xnnext);
-		if (NULL != xnnext->match)
+		assert(xnnext != NULL);
+		if (xnnext->match != NULL)
 			continue;
 
-		if (NULL == (mnext = TAILQ_NEXT(mchild, entries)))
+		if ((mnext = TAILQ_NEXT(mchild, entries)) == NULL)
 			continue;
 		if (TAILQ_EMPTY(&mnext->children))
 			continue;
 		xmnext = &oldmap->nodes[mnext->id];
-		assert(NULL != xmnext);
-		if (NULL != xmnext->match)
+		assert(xmnext != NULL);
+		if (xmnext->match != NULL)
 			continue;
 
 		if ( ! match_eq(nnext, mnext))
 			continue;
 
-#if DEBUG
-		warnx("%s (%zu): subtree match: %s (%zu) -> %zu",
-			names[n->type], n->id, 
-			names[nchild->type], nchild->id, mchild->id);
-#endif
-
 		xnnext->match = mnext;
 		xmnext->match = nnext;
 	}
@@ -1205,8 +1223,8 @@ static void
 node_optimise_bottomup(const struct lowdown_node *n, 
 	struct xmap *newmap, struct xmap *oldmap)
 {
-	const struct lowdown_node *nn, *on, *nnn, *maxn = NULL;
-	double		w, maxw = 0.0, tw = 0.0;
+	const struct lowdown_node	*nn, *on, *nnn, *maxn = NULL;
+	double				 w, maxw = 0.0, tw = 0.0;
 
 	/* Do a depth-first pre-order search. */
 
@@ -1223,17 +1241,17 @@ node_optimise_bottomup(const struct lowdown_node *n,
 	 * If we're already matched, then move on.
 	 */
 
-	if (NULL != newmap->nodes[n->id].match)
+	if (newmap->nodes[n->id].match != NULL)
 		return;
 
 	TAILQ_FOREACH(nn, &n->children, entries) {
-		if (NULL == newmap->nodes[nn->id].match)
+		if (newmap->nodes[nn->id].match == NULL)
 			continue;
-		if (NULL == (on = newmap->nodes[nn->id].match->parent))
+		if ((on = newmap->nodes[nn->id].match->parent) == NULL)
 			continue;
 		if (on == maxn)
 			continue;
-		if ( ! match_eq(n, on))
+		if (!match_eq(n, on))
 			continue;
 		
 		/*
@@ -1247,7 +1265,7 @@ node_optimise_bottomup(const struct lowdown_node *n,
 
 		w = 0.0;
 		TAILQ_FOREACH(nnn, &n->children, entries) {
-			if (NULL == newmap->nodes[nnn->id].match)
+			if (newmap->nodes[nnn->id].match == NULL)
 				continue;
 			if (on != newmap->nodes[nnn->id].match->parent)
 				continue;
@@ -1264,7 +1282,7 @@ node_optimise_bottomup(const struct lowdown_node *n,
 
 	/* See if we found any similar sub-trees. */
 
-	if (NULL == maxn)
+	if (maxn == NULL)
 		return;
 
 	/*
@@ -1283,13 +1301,13 @@ struct lowdown_node *
 lowdown_diff(const struct lowdown_node *nold,
 	const struct lowdown_node *nnew, size_t *maxn)
 {
-	struct xmap	 xoldmap, xnewmap;
-	struct xnode	*xnew, *xold;
-	struct pnodeq	 pq;
-	struct pnode	*p;
-	const struct lowdown_node *n, *nn;
-	struct lowdown_node *comp;
-	size_t		 i;
+	struct xmap			 xoldmap, xnewmap;
+	struct xnode			*xnew, *xold;
+	struct pnodeq			 pq;
+	struct pnode			*p;
+	const struct lowdown_node	*n, *nn;
+	struct lowdown_node		*comp = NULL;
+	size_t				 i;
 
 	memset(&xoldmap, 0, sizeof(struct xmap));
 	memset(&xnewmap, 0, sizeof(struct xmap));
@@ -1301,12 +1319,15 @@ lowdown_diff(const struct lowdown_node *nold,
 	 * See "Phase 2", sec 5.2.
 	 */
 
-	(void)assign_sigs(NULL, &xoldmap, nold);
-	(void)assign_sigs(NULL, &xnewmap, nnew);
+	if (assign_sigs(NULL, &xoldmap, nold) < 0.0)
+		goto out;
+	if (assign_sigs(NULL, &xnewmap, nnew) < 0.0)
+		goto out;
 
 	/* Prime the priority queue with the root. */
 
-	pqueue(nnew, &xnewmap, &pq);
+	if (!pqueue(nnew, &xnewmap, &pq))
+		goto out;
 
 	/* 
 	 * Match-make while we have nodes in the priority queue.
@@ -1314,16 +1335,15 @@ lowdown_diff(const struct lowdown_node *nold,
 	 * See "Phase 3", sec 5.2.
 	 */
 
-	while (NULL != (p = TAILQ_FIRST(&pq))) {
-		/* TODO: cache of pnodes. */
+	while ((p = TAILQ_FIRST(&pq)) != NULL) {
 		TAILQ_REMOVE(&pq, p, entries);
 		n = p->node;
 		free(p);
 
 		xnew = &xnewmap.nodes[n->id];
-		assert(NULL == xnew->match);
-		assert(NULL == xnew->optmatch);
-		assert(0 == xnew->opt);
+		assert(xnew->match == NULL);
+		assert(xnew->optmatch == NULL);
+		assert(xnew->opt == 0);
 
 		/*
 		 * Look for candidates: if we have a matching signature,
@@ -1334,22 +1354,23 @@ lowdown_diff(const struct lowdown_node *nold,
 
 		for (i = 0; i < xoldmap.maxid + 1; i++) {
 			xold = &xoldmap.nodes[i];
-			if (NULL == xold->node)
+			if (xold->node == NULL)
 				continue;
-			if (NULL != xold->match)
+			if (xold->match != NULL)
 				continue;
 			if (strcmp(xnew->sig, xold->sig))
 				continue;
 
-			assert(NULL == xold->match);
+			assert(xold->match == NULL);
 			candidate(xnew, &xnewmap, xold, &xoldmap);
 		}
 
 		/* No match: enqueue children ("Phase 3" cont.). */
 
-		if (NULL == xnew->optmatch) {
+		if (xnew->optmatch == NULL) {
 			TAILQ_FOREACH(nn, &n->children, entries)
-				pqueue(nn, &xnewmap, &pq);
+				if (!pqueue(nn, &xnewmap, &pq))
+					goto out;
 			continue;
 		}
 
@@ -1360,8 +1381,8 @@ lowdown_diff(const struct lowdown_node *nold,
 		 * See "Phase 3", sec. 5.2.
 		 */
 
-		assert(NULL == xnew->match);
-		assert(NULL == xoldmap.nodes[xnew->optmatch->id].match);
+		assert(xnew->match == NULL);
+		assert(xoldmap.nodes[xnew->optmatch->id].match == NULL);
 
 		match_down(xnew, &xnewmap, 
 			&xoldmap.nodes[xnew->optmatch->id], &xoldmap);
@@ -1378,11 +1399,6 @@ lowdown_diff(const struct lowdown_node *nold,
 	node_optimise_topdown(nnew, &xnewmap, &xoldmap);
 	node_optimise_bottomup(nnew, &xnewmap, &xoldmap);
 
-#if DEBUG
-	node_print(nnew, &xnewmap, 0);
-	node_print(nold, &xoldmap, 0);
-#endif
-
 	/*
 	 * The tree is optimal.
 	 * Now we need to compute the delta and merge the trees.
@@ -1396,10 +1412,13 @@ lowdown_diff(const struct lowdown_node *nold,
 		xnewmap.maxid + 1 :
 		xoldmap.maxid + 1;
 
-	/* Clean up and exit. */
-
-	assert(TAILQ_EMPTY(&pq));
+out:
+	assert(comp != NULL);
+	while ((p = TAILQ_FIRST(&pq)) != NULL) {
+		TAILQ_REMOVE(&pq, p, entries);
+		free(p);
+	}
 	free(xoldmap.nodes);
 	free(xnewmap.nodes);
-	return(comp);
+	return comp;
 }
